@@ -13,7 +13,6 @@ pragma to the header of your file
 
 -}
 
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE TupleSections #-}
 {-# LANGUAGE TemplateHaskellQuotes #-}
 
@@ -26,7 +25,7 @@ where
 -- external
 import Control.Monad.Trans.Maybe (MaybeT (..))
 import Data.Maybe (catMaybes)
-import GHC.TcPluginM.Extra (evByFiat, tracePlugin, newWanted)
+import GHC.TcPluginM.Extra (evByFiatWithDependencies, tracePlugin, newWanted)
 import qualified Data.Type.Ord
 import qualified GHC.TypeError
 
@@ -35,31 +34,26 @@ import GHC.Builtin.Names (eqPrimTyConKey, hasKey, getUnique)
 import GHC.Builtin.Types (promotedTrueDataCon, promotedFalseDataCon)
 import GHC.Builtin.Types (boolTy, naturalTy, cTupleDataCon, cTupleTyCon)
 import GHC.Builtin.Types.Literals (typeNatDivTyCon, typeNatModTyCon, typeNatCmpTyCon)
-import GHC.Core.Coercion (mkUnivCo)
+import GHC.Core.Coercion (Coercion, mkUnivCo)
 import GHC.Core.DataCon (dataConWrapId)
 import GHC.Core.Predicate (EqRel (NomEq), Pred (EqPred, IrredPred), classifyPredType)
 import GHC.Core.Reduction (Reduction(..))
 import GHC.Core.TyCon (TyCon)
 import GHC.Core.TyCo.Rep (Type (..), TyLit (..), UnivCoProvenance (PluginProv))
 import GHC.Core.Type (Kind, mkTyConApp, splitTyConApp_maybe, typeKind)
-#if MIN_VERSION_ghc(9,6,0)
 import GHC.Core.TyCo.Compare (eqType)
-#else
-import GHC.Core.Type (eqType)
-#endif
 import GHC.Data.IOEnv (getEnv)
 import GHC.Driver.Env (hsc_NC)
 import GHC.Driver.Plugins (Plugin (..), defaultPlugin, purePlugin)
 import GHC.Plugins (thNameToGhcNameIO)
-import GHC.Tc.Plugin (TcPluginM, tcLookupTyCon, tcPluginTrace, tcPluginIO, unsafeTcPluginTcM)
-import GHC.Tc.Types (TcPlugin(..), TcPluginSolveResult (..), TcPluginRewriter, TcPluginRewriteResult (..), Env (env_top))
+import GHC.Tc.Plugin
+  (TcPluginM, tcLookupTyCon, tcPluginTrace, tcPluginIO, unsafeTcPluginTcM)
+import GHC.Tc.Types
+  (TcPlugin(..), TcPluginSolveResult (..), TcPluginRewriter, TcPluginRewriteResult (..),
+   Env (env_top))
+import GHC.Tc.Types.Constraint (Ct, ctEvidence, ctEvPred, ctLoc, isWantedCt)
 import GHC.Tc.Types.Constraint
-  (Ct, ctEvidence, ctEvPred, ctLoc, isWantedCt)
-#if MIN_VERSION_ghc(9,8,0)
-import GHC.Tc.Types.Constraint (Ct (..), DictCt(..), EqCt(..), IrredCt(..), qci_ev)
-#else
-import GHC.Tc.Types.Constraint (Ct (CQuantCan), qci_ev, cc_ev)
-#endif
+  (Ct (..), DictCt(..), EqCt(..), IrredCt(..), ctEvCoercion, qci_ev)
 import GHC.Tc.Types.Evidence (EvTerm, EvBindsVar, Role(..), evCast, evId)
 import GHC.Types.Unique.FM (UniqFM, listToUFM)
 import GHC.Utils.Outputable (Outputable (..), (<+>), ($$), text)
@@ -130,7 +124,7 @@ extraRewrite defs = listToUFM
 
     reduce tc args res = Reduction co res
      where
-      co = mkUnivCo (PluginProv "ghc-typelits-extra") Nominal
+      co = mkUnivCo (PluginProv "ghc-typelits-extra") [] Nominal
              (mkTyConApp tc args) res
 
 
@@ -150,19 +144,23 @@ decideEqualSOP defs _ givens  wanteds = do
 
 data SolverConstraint
    = NatEquality Ct ExtraOp ExtraOp Normalised
-   | NatInequality Ct ExtraOp ExtraOp Bool Normalised
+   | NatInequality Ct [Coercion] ExtraOp ExtraOp Bool Normalised
 
 instance Outputable SolverConstraint where
-  ppr (NatEquality ct op1 op2 norm) = text "NatEquality" $$ ppr ct $$ ppr op1 $$ ppr op2 $$ ppr norm
-  ppr (NatInequality _ op1 op2 b norm) = text "NatInequality" $$ ppr op1 $$ ppr op2 $$ ppr b $$ ppr norm
+  ppr (NatEquality ct op1 op2 norm) =
+    text "NatEquality" $$ ppr ct $$ ppr op1 $$ ppr op2 $$ ppr norm
+  ppr (NatInequality _ _ op1 op2 b norm) =
+    text "NatInequality" $$ ppr op1 $$ ppr op2 $$ ppr b $$ ppr norm
 
 data SimplifyResult
   = Simplified [(EvTerm,Ct)] [Ct]
   | Impossible SolverConstraint
 
 instance Outputable SimplifyResult where
-  ppr (Simplified evs new) = text "Simplified" $$ text "Solved:" $$ ppr evs $$ text "New:" $$ ppr new
-  ppr (Impossible sct)  = text "Impossible" <+> ppr sct
+  ppr (Simplified evs new) =
+    text "Simplified" $$ text "Solved:" $$ ppr evs $$ text "New:" $$ ppr new
+  ppr (Impossible sct) =
+    text "Impossible" <+> ppr sct
 
 simplifyExtra :: ExtraDefs -> [SolverConstraint] -> TcPluginM SimplifyResult
 simplifyExtra defs eqs = tcPluginTrace "simplifyExtra" (ppr eqs) >> simples [] [] eqs
@@ -173,45 +171,48 @@ simplifyExtra defs eqs = tcPluginTrace "simplifyExtra" (ppr eqs) >> simples [] [
       ur <- unifyExtra ct u v
       tcPluginTrace "unifyExtra result" (ppr ur)
       case ur of
-        Win                          -> simples (((,) <$> evMagic ct <*> pure ct):evs) news eqs'
+        Win -> simples (((,) <$> evMagic ct [] <*> pure ct):evs) news eqs'
         Lose | null evs && null eqs' -> return (Impossible eq)
         _ | norm == Normalised && isWantedCt ct -> do
           newCt <- createWantedFromNormalised defs eq
-          simples (((,) <$> evMagic ct <*> pure ct):evs) (newCt:news) eqs'
+          simples (((,) <$> evMagic ct [] <*> pure ct):evs) (newCt:news) eqs'
         Lose -> simples evs news eqs'
         Draw -> simples evs news eqs'
-    simples evs news (eq@(NatInequality ct u v b norm):eqs') = do
+    simples evs news (eq@(NatInequality ct deps u v b norm):eqs') = do
       tcPluginTrace "unifyExtra leq result" (ppr (u,v,b))
       case (u,v) of
         (I i,I j)
-          | (i <= j) == b -> simples (((,) <$> evMagic ct <*> pure ct):evs) news eqs'
+          | (i <= j) == b -> simples (((,) <$> evMagic ct deps <*> pure ct):evs) news eqs'
           | otherwise     -> return  (Impossible eq)
         (p, Max x y)
-          | b && (p == x || p == y) -> simples (((,) <$> evMagic ct <*> pure ct):evs) news eqs'
+          | b && (p == x || p == y)
+          -> simples (((,) <$> evMagic ct deps <*> pure ct):evs) news eqs'
 
         -- transform:  q ~ Max x y => (p <=? q ~ True)
         -- to:         (p <=? Max x y) ~ True
         -- and try to solve that along with the rest of the eqs'
         (p, q@(V _))
           | b -> case findMax q eqs of
-                   Just m  -> simples evs news (NatInequality ct p m b norm:eqs')
+                   Just (i,m) ->
+                      simples evs news
+                        (NatInequality ct (i:deps) p m b norm:eqs')
                    Nothing -> simples evs news eqs'
         _ | norm == Normalised && isWantedCt ct -> do
           newCt <- createWantedFromNormalised defs eq
-          simples (((,) <$> evMagic ct <*> pure ct):evs) (newCt:news) eqs'
+          simples (((,) <$> evMagic ct deps <*> pure ct):evs) (newCt:news) eqs'
         _ -> simples evs news eqs'
 
     -- look for given constraint with the form: c ~ Max x y
-    findMax :: ExtraOp -> [SolverConstraint] -> Maybe ExtraOp
+    findMax :: ExtraOp -> [SolverConstraint] -> Maybe (Coercion, ExtraOp)
     findMax c = go
       where
         go [] = Nothing
         go ((NatEquality ct a b@(Max _ _) _) :_)
           | c == a && not (isWantedCt ct)
-            = Just b
+            = Just (ctEvCoercion (ctEvidence ct), b)
         go ((NatEquality ct a@(Max _ _) b _) :_)
           | c == b && not (isWantedCt ct)
-            = Just a
+            = Just (ctEvCoercion (ctEvidence ct), a)
         go (_:rest) = go rest
 
 
@@ -235,8 +236,12 @@ toSolverConstraint defs ct = case classifyPredType $ ctEvPred $ ctEvidence ct of
       -> do
           (x', n1) <- normaliseNat defs x
           (y', n2) <- normaliseNat defs y
-          let res | tc' == promotedTrueDataCon  = pure (NatInequality ct x' y' True  (mergeNormalised n1 n2))
-                  | tc' == promotedFalseDataCon = pure (NatInequality ct x' y' False (mergeNormalised n1 n2))
+          let res | tc' == promotedTrueDataCon
+                  = pure (NatInequality ct [] x' y' True
+                            (mergeNormalised n1 n2))
+                  | tc' == promotedFalseDataCon
+                  = pure (NatInequality ct [] x' y' False
+                            (mergeNormalised n1 n2))
                   | otherwise                   = fail "Nothing"
           res
       | TyConApp tc [TyConApp ordCondTc zs, _] <- t1
@@ -256,7 +261,7 @@ toSolverConstraint defs ct = case classifyPredType $ ctEvPred $ ctEvidence ct of
       -> do
           (x', n1) <- normaliseNat defs x
           (y', n2) <- normaliseNat defs y
-          pure (NatInequality ct x' y' True  (mergeNormalised n1 n2))
+          pure (NatInequality ct [] x' y' True  (mergeNormalised n1 n2))
     IrredPred (TyConApp tc [TyConApp ordCondTc zs, _])
       | tc == assertTC defs
       , ordCondTc == ordTyCon defs
@@ -272,7 +277,7 @@ toSolverConstraint defs ct = case classifyPredType $ ctEvPred $ ctEvidence ct of
       -> do
           (x', n1) <- normaliseNat defs x
           (y', n2) <- normaliseNat defs y
-          pure (NatInequality ct x' y' True  (mergeNormalised n1 n2))
+          pure (NatInequality ct [] x' y' True  (mergeNormalised n1 n2))
     _ -> fail "Nothing"
   where
     isNatKind :: Kind -> Bool
@@ -281,7 +286,7 @@ toSolverConstraint defs ct = case classifyPredType $ ctEvPred $ ctEvidence ct of
 createWantedFromNormalised :: ExtraDefs -> SolverConstraint -> TcPluginM Ct
 createWantedFromNormalised defs sct = do
   let extractCtSides (NatEquality ct t1 t2 _)   = (ct, reifyEOP defs t1, reifyEOP defs t2)
-      extractCtSides (NatInequality ct x y b _) =
+      extractCtSides (NatInequality ct _ x y b _) =
         let tc = if b then promotedTrueDataCon else promotedFalseDataCon
             t1 = TyConApp (ordTyCon defs)
                     [ boolTy
@@ -300,19 +305,15 @@ createWantedFromNormalised defs sct = do
   ev <- newWanted (ctLoc ct) newPredTy
   let ctN = case ct of
               CQuantCan qc -> CQuantCan (qc { qci_ev = ev})
-#if MIN_VERSION_ghc(9,8,0)
               CDictCan di     -> CDictCan (di { di_ev = ev})
               CIrredCan ir    -> CIrredCan (ir { ir_ev = ev})
               CEqCan eq       -> CEqCan (eq { eq_ev = ev})
               CNonCanonical _ -> CNonCanonical ev
-#else
-              ctX -> ctX { cc_ev = ev }
-#endif
   return ctN
 
 fromSolverConstraint :: SolverConstraint -> Ct
 fromSolverConstraint (NatEquality ct _ _ _)  = ct
-fromSolverConstraint (NatInequality ct _ _ _ _) = ct
+fromSolverConstraint (NatInequality ct _ _ _ _ _) = ct
 
 lookupExtraDefs :: TcPluginM ExtraDefs
 lookupExtraDefs = do
@@ -337,12 +338,12 @@ lookupTHName th = do
     maybe (fail $ "Failed to lookup " ++ show th) return res
 
 -- Utils
-evMagic :: Ct -> Maybe EvTerm
-evMagic ct = case classifyPredType $ ctEvPred $ ctEvidence ct of
-    EqPred NomEq t1 t2 -> Just (evByFiat "ghc-typelits-extra" t1 t2)
+evMagic :: Ct -> [Coercion] -> Maybe EvTerm
+evMagic ct deps = case classifyPredType $ ctEvPred $ ctEvidence ct of
+    EqPred NomEq t1 t2 -> Just (evByFiatWithDependencies "ghc-typelits-extra" deps t1 t2)
     IrredPred p ->
       let t1 = mkTyConApp (cTupleTyCon 0) []
-          co = mkUnivCo (PluginProv "ghc-typelits-extra") Representational t1 p
+          co = mkUnivCo (PluginProv "ghc-typelits-extra") deps Representational t1 p
           dcApp = evId (dataConWrapId (cTupleDataCon 0))
        in Just (evCast dcApp co)
     _ -> Nothing
